/*
 *******************************************************
  
 cat an mice benchmark, wodes08, libfaudes 2.09

 This file performs the locally modular supervisor synthesis according to

 M. H. de Queiroz and J. E. R. Cury, Modular supervisory control
 of large scale discrete event systems, Workshop on Discrete Event
 Systems, 2000.

 Each cat/mouse is considered as an individual subplant such that the 
 different subplants to not share any events. A modular synthesis is carried 
 out for each cat/mouse specification: 
 - for the k-th cat specification, the k-th cat composed with all mice forms the plant
 - for the k-th mouse specification, the k-th mouse composed with all cats forms the plant
 
 *******************************************************
 */ 



#include <stdio.h>
#include <iostream>
#include "libfaudes.h"
#include "catsMice_individual_models.h"

using namespace std;
using namespace faudes;

// locally modular design
void LocallyModularCat_and_Mouse(const int n, const int k){
    FD_DF("LocallyModular_Cat_and_Mouse(" << n << "," << k << ")");
    // helper data structures
    cGenerator cat,mouse;
    cGenerator catSpec, mouseSpec;
    cGenerator spec, plant, sup, tmp, tmpSpec;
    vector<pair<cGenerator, cGenerator> > catVec;
    catVec = vector<pair<cGenerator, cGenerator> >(k);
    vector<pair<cGenerator, cGenerator> > mouseVec;
    mouseVec = vector<pair<cGenerator, cGenerator> >(k);    
    
    vector<cGenerator> mouseSupVec;
    mouseSupVec = vector<cGenerator>(k);
    vector<cGenerator> catSupVec;
    catSupVec = vector<cGenerator>(k);
    
    unsigned int statesSum = 0;
    unsigned int transitionsSum = 0;
    // compute k cat/mouse specifications and compute the parallel composition of all cats/mice
    for( int j = 1; j <= k; j++){
        CatSpecification(catVec[j-1].first,catVec[j-1].second,j,n,k);
        MouseSpecification(mouseVec[j-1].first,mouseVec[j-1].second,j,n,k);
        if(j == 1){
            cat = catVec[j-1].first;    
            mouse = mouseVec[j-1].first;
        }
        else{
             tmp = cat;
             cParallel(tmp,catVec[j-1].first,cat);
             tmp = mouse;
             cParallel(tmp,mouseVec[j-1].first,mouse);
             tmp.Clear();
        }        
    }
    //  the result up to now is a cat plant, a mouse plant, and pairs of cats/mice with respective modular specifications
    // Next, the locally modular supervisors for the local plants are computed 
    for( int j = 1; j <= k; j++){
        // modular supervisor for the j-th cat with all mice
        cParallel(catVec[j-1].first,mouse,plant);
        cSupConNB(plant,catVec[j-1].second,catSupVec[j-1]);
        catSupVec[j-1].ClearStateNames();
        statesSum += catSupVec[j-1].Size();
        transitionsSum += catSupVec[j-1].TransRel().Size();
        // modular supervisor for the j-th mouse with all cats
        cParallel(mouseVec[j-1].first,cat,plant);
        cSupConNB(plant,mouseVec[j-1].second,mouseSupVec[j-1]);  
        mouseSupVec[j-1].ClearStateNames();
        statesSum += mouseSupVec[j-1].Size();   
        transitionsSum += mouseSupVec[j-1].TransRel().Size();
    }
    // compute the overall closed-loop for verification of nonconflict:
    // initialization with first cat and mouse
    cParallel(catSupVec[0],mouseSupVec[0],plant);
    plant.ClearStateNames();
    // parallel composition of the remaining cats/mice
    for( int j= 2; j <= k; j++){
         cParallel(plant,catSupVec[j-1],tmp);
         cParallel(tmp,mouseSupVec[j-1],plant);     
    }
    plant.ClearStateNames();
    std::cout << "#############################################################\n";
    std::cout << "#Result of Locally Modular Synthesis for n = " << n << " and k = " << k << endl;
    if(plant.IsTrim() ){
        std::cout << "modular supervisors are nonconflicting! " 
        << " closed loop size: " << plant.Size() << ", transitions: " << plant.TransRel().Size() 
        << ", supervisor size: " << statesSum << ", transitions: " << transitionsSum << endl; 
        // write the modular supervisors to file output
        for( int j = 1; j <= k; j++){
            catSupVec[j-1].Write( ("./results/locally_modular/catSup" + ToStringInteger(j) + ".gen").c_str() );
            mouseSupVec[j-1].Write( ("./results/locally_modular/mouseSup" + ToStringInteger(j) + ".gen").c_str() );
        }
    }
    else{
        cSupConNB(plant,plant,sup);
        sup.ClearStateNames();
        std::cout << "modular supervisors are conflicting! " 
        << " plant size: " << plant.Size() << ", transitions: " << plant.TransRel().Size() 
        << ", closed loop size: " << sup.Size() << ", transitions: " << sup.TransRel().Size() << endl;
        // monolithic supervisor can be written to file output
        sup.Write("./results/locally_modular/CatMouseSup.gen");   
    }
    std::cout << "#############################################################\n";
}

// doit
int main(int argc, char *argv[]){
    try{
        // command line parsing; two input parameters have to be provided
        if(argc!=3) {
            cerr<<"usage: ./catsMice_individual_monolithic <n> <k> "<<endl;
            return(-1);
        }
        // number of levels
        int n=atoi(argv[1]);
        // number of cats and mice
        int k=atoi(argv[2]);
        // locally modular supervisor synthesis
        LocallyModularCat_and_Mouse(n,k);
       
    } //endof try
    catch (Exception& e) {
        std::cout << "function: " << e.Where() << std::endl;
        std::cout << "exception descrition: " << e.What() << std::endl;
        std::cout << "exception id: " << e.Id() << std::endl;
        return(-1);
    }
    // return 0 after correct program execution
    return 0;
}

