/*
 *******************************************************
  
 cats an mice benchmark, wodes08, libfaudes 2.09

 This file preforms the hierarchical and decentralized supervisor synthesis for the 
 cat-and-mouse benchmark example, where individual maze levels are modeled. The synthesis 
 approach is developed in 
 - K. Schmidt. (2005) Hierarchical control of decentralized discrete event
   systems: Theory and application. Phd-thesis, Lehrstuhl f ur Regelungstechnik,
   Universitat Erlangen-Nurnberg. [Online]. Available:
   http://www.rt.eei.uni-erlangen.de/FGdes/dissertation2005 schmidt.pdf
 and extended to maximally permissive control in
 - K. Schmidt and C. Breindl, On maximal permissiveness of hierarchical
   and modular supervisory control approaches for discrete event
   systems, Workshop on Discrete Event Systems (WODES), 2008. 
 
 *******************************************************
 */ 


#include <stdio.h>
#include <iostream>
#include <cmath>
#include "libfaudes.h"
#include "op_observer.h"
#include "catsMice_levels_models.h"

using namespace std;
using namespace faudes;


// HierarchicalMultiLevelCat_and_Mouse(n, k) 
void HierarchicalMultiLevelCat_and_Mouse(const int n, const int k){
     std::string nstr,kstr;
    nstr = ToStringInteger(n);
    kstr = ToStringInteger(k);     
    // cGenerators and high-level alphabet for intermediate computations 
    cGenerator maze, mazeSpec,sup, tmp, supHigh, plant;
    EventSet highAlphabet;
    // vector for the locally controlled mazes in each level
    vector<cGenerator> supVec(n);
    // vector for the initial high-level alphabets for each maze level
    vector<cEventSet> highAlphVec(n);
    // data structurs for the relabeling process after the observer synthesis
    map<Transition,Idx> mapRelabeledTrans; // returned by calcAbstAlph
    map<Idx,set<Idx> > relabeledEventsMap; // events to be relabeled in neighboring maze levels
    map<Idx,set<Idx> >::const_iterator reIt, reEndIt; 
    TransSet::Iterator tsIt, tsEndIt;
    map<Transition,Idx>::const_iterator rtIt, rtEndIt;    
    set<Idx>::const_iterator rsIt, rsEndIt;
    vector<cGenerator> allSupVec;
    // ===========================================
    // main part of the hierarchical construction:
    // ===========================================
    // compute the local supervisors iteratively starting from the lowest level:
    // - plant and specification are determined
    // - local supervisor is computed
    // - abstraction alphabet is computed using the observer algorithm including local control consistency (LCC)
    // - events in neighboring maze levels are relabeled and the corresponding 
    //   transitions are added
    // - two vectors with the resulting relabeled local supervisors and the abstraction alphabets 
    //   are returned
    // ===========================================
    
    // Local supervisor computation for all maze levels (first level of the supervisor hierarchy)
    for(int l = 1; l <= n; l++){
        highAlphabet.Clear();
        // computation of the model for the current maze level
        Maze(maze,mazeSpec,n,k,l,highAlphabet);
        maze.ClearStateNames();
        mazeSpec.ClearStateNames();
        // computation of the local supervisors
        cSupConNB(maze,mazeSpec,supVec[l-1]);
        supVec[l-1].ClearStateNames();
        // update the current automaton with the events/transitions changed in the abstraction of the previous level
        // this step is not executed for the first maze level as there is no predecessor. 
        if(l > 1){
          insertRelabeledEvents(supVec[l-1],relabeledEventsMap,highAlphVec[l-1]);
        }  
        // compute the abstraction for the current level (fulfills Lm-obsever property and LCC)
	    relabeledEventsMap.clear();
        calcAbstAlphObsLCC(supVec[l-1], highAlphabet, highAlphVec[l-1], relabeledEventsMap);
        supVec[l-1].ClearStateNames();
        // update the lower-level automaton with the changed events/transitions 
        if(l > 1){
          insertRelabeledEvents(supVec[l-2],relabeledEventsMap,highAlphVec[l-2]);
        }
        // go to the next level of the maze
    } 
    
    // After all local supervisors have been determined, the supervisor hierarchy is constructed. 
    // Each high-level plant is computed by parallel composition of two abstracted local supervisors.
    // Then, the resulting high-level plants are used as local plants for the next synthesis step. Iterative
    // hierarchical abstraction and local synthesis are performed until there is a single plant on the 
    // highest level of the supervisor hierarchy. 
    int loopCounter = n;
    vector<cGenerator> plantVec;
    vector<cEventSet> highItAlphVec;
    cGenerator high1, high2;
    int highStatesSum = 0;
    int transitionsSum = 0;
    int level = 1;
    while(1){
       level++;
        bool breakLoop = false;
        loopCounter = (int) ceil((float)loopCounter/2); // in each hierarchical step, the number of local models is halved
        plantVec = vector<cGenerator>(loopCounter);
        highItAlphVec = vector<cEventSet>(loopCounter);
        for(int x = 1; x <= loopCounter; x++){  
            highStatesSum = highStatesSum + supVec[(x-1)*2].Size();
            transitionsSum = transitionsSum + supVec[(x-1)*2].TransRel().Size();  
            // if two maze levels can be grouped, compute the grouped plant and the corresponding abstraction alphabet
            if((int) supVec.size() >= x*2){
                highStatesSum = highStatesSum + supVec[(x-1)*2+1].Size();
                transitionsSum = transitionsSum + supVec[(x-1)*2+1].TransRel().Size();
                cProject(supVec[(x-1)*2], highAlphVec[(x-1)*2],high1);
                cProject(supVec[(x-1)*2+1], highAlphVec[(x-1)*2+1], high2);
                high1.ClearStateNames();
                high2.ClearStateNames();
                // composition of two abstracted plants
                cParallel( high1, high2, plant);
                plant.ClearStateNames();
                // local nonblocking synthesis for the current high-level plant
                cSupConNB(plant,plant,plantVec[x-1]);
                plantVec[x-1].ClearStateNames();
                // the initial high-level alphabet for the next abstraction step is computed from the shared events
                // with the neighboring plants
                if(x > 1 && x < loopCounter){// two neighboring plants
                     highItAlphVec[x-1] = highItAlphVec[x-1] + (supVec[(x-1)*2-1].Alphabet()*supVec[(x-1)*2].Alphabet() );
                     highItAlphVec[x-1] = highItAlphVec[x-1] + (supVec[(x-1)*2+1].Alphabet()*supVec[(x-1)*2+2].Alphabet() );
                }
                else if(x == 1 && x < loopCounter){// one right neighbor
                    highItAlphVec[x-1] = highItAlphVec[x-1] + (supVec[(x-1)*2+1].Alphabet()*supVec[(x-1)*2+2].Alphabet() );
                }
                else if( x == loopCounter && x > 1){// one left neighbor
                    highItAlphVec[x-1] = highItAlphVec[x-1] + (supVec[(x-1)*2-1].Alphabet()*supVec[(x-1)*2].Alphabet() );
                }
                allSupVec.push_back(supVec[(x-1)*2]);
                allSupVec.push_back(supVec[(x-1)*2+1]);
                if(supVec.size() == 2){
                    allSupVec.push_back(plantVec[x-1]);
                    breakLoop = true;
                }
            }
            // if only one plant is left, keep this plant - no more abstraction is possible
            else{
                 plantVec[x-1] = supVec[(x-1)*2];
                 highItAlphVec[x-1] = supVec[(x-1)*2].Alphabet();
                 supVec[(x-1)*2].Clear();  
                 if(supVec.size() == 1)
                     breakLoop = true;
            }
        }
        // break the while-loop if the highest-level supervisor is reached
        if(breakLoop)
            break;
            
        // now we have a vector of new low-level plants (plantVec) and the corresponding high-level alphabets for the next abstraction step (highItAlphVec)
        // in the next step, the abstraction is carried out
        // the new set of low-level supervisors is set to the computed plant vector
        supVec.clear();
        supVec.resize(loopCounter);
        supVec = plantVec;
        plantVec.clear();
        for(int x = 1; x <= loopCounter; x++){
            if(loopCounter == 1)
                break;
            highAlphabet.Clear();
            // update the current automaton with the events/transitions changed in the abstraction of the previous level
            if(x > 1){
               insertRelabeledEvents(supVec[x-1],relabeledEventsMap,highItAlphVec[x-1]);
            }
            // compute the abstraction for the current level
            EventSet contrEvs=supVec[x-1].ControllableEvents();
	        relabeledEventsMap.clear();
            calcAbstAlphObsLCC(supVec[x-1], highItAlphVec[x-1], highAlphVec[x-1], relabeledEventsMap);
            // insert all transitions with relabeled events into previously computed supervisors (allSupVec)
            for(int v = 0; v < (int) allSupVec.size();  v++){
               insertRelabeledEvents(allSupVec[v],relabeledEventsMap);
	       }
            // update the lower-level automaton with the changed events/transitions 
            if(x > 1){
              insertRelabeledEvents(supVec[x-2],relabeledEventsMap,highAlphVec[x-2]);
            }
            // now we have the locel plants (supVec) and their next abstraction alphabets (highAlphVec)    
        }
 
    }
    std::cout << "#############################################################\n";
    std::cout << "#Result of Hierarchical Multilevel Synthesis for n = " << n << " and k = " << k << endl;
    std::cout << " overall supervisor states: " << highStatesSum << " overall supervisor transitions: " << transitionsSum << endl;
    std::cout << "#############################################################\n";
    vector<cGenerator>::const_iterator allIt;
    for(int i = 0; i < allSupVec.size(); i++){
        string filename = "./results/multilevel/sup" + ToStringInteger(i) + ".gen";
        allSupVec[i].Write(filename.c_str() );
    }
}


// main function for the hierarchical multilevel synthesis
int main(int argc, char* argv[]) {
    try{
        // command line parsing; two input parameters have to be provided
        if(argc!=3) {
            cerr<<"usage: ./catsMice_individual_monolithic <n> <k> "<<endl;
            return(-1);
        }
        // number of levels
        int n=atoi(argv[1]);
        // number of cats and mice
        int k=atoi(argv[2]);
        // Perform the monolithic supervisor synthesis
        HierarchicalMultiLevelCat_and_Mouse(n,k);
    } //endof try
    catch (Exception& e) {
        std::cout << "function: " << e.Where() << std::endl;
        std::cout << "exception descrition: " << e.What() << std::endl;
        std::cout << "exception id: " << e.Id() << std::endl;
        return(-1);
    }
    // return 0 after correct program execution
    return 0;
}


