/** @file vgenerator.h Class vGenerator */

/* FAU Discrete Event Systems Library (libfaudes)

Copyright (C) 2006  Bernd Opitz
Copyright (C) 2007  Thomas Moor
Exclusive copyright is granted to Klaus Schmidt

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNES FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef VGENERATOR_H

#include "definitions.h"
#include "exception.h"
#include "symboltable.h"
#include "indexset.h"
#include "nameset.h"
#include "transset.h"
#include "token.h"
#include "tokenreader.h"
#include "tokenwriter.h"
#include <map>
#include <set>
#include <sstream>
#include <cstdlib>
#include <assert.h>

namespace faudes {

/**
 * Virtual version of a plain generator (ie no attributes). This class is the
 * common base of all generator classes. It defines the application interface
 * for a plain finite state machine with no attributes and is meant as argument
 * and result type of functions that implement general purpose operations on formal
 * languages (Projection, Kleene closure, minimal Realisation etc). 
 *
 * Technically, the vGenerator class is a virtual class and only implements elementary
 * maintenance of generators (unique id, generator name, symboltables etc). It does not implement 
 * any operations related to states, events and transitions. The latter is done in 
 * TaGenerator, where template parameters of AttributeVoid effectively model 
 * a plain finite state machine without any attributes.
 *
 *
 * \ingroup GeneratorClasses
 */

class vGenerator {
 public:
    
  /** @name Constructors & Destructor */
  /** @{ doxygen group */

  /**
   * Default constructor
   */
  vGenerator(void);


  /**
   * Copy-constructror
   */
  vGenerator(const vGenerator& rOtherGen);

  /**
   * Destructor 
   */
  virtual ~vGenerator(void) {};

  /**
   * Construct on heap. Derived classes must reimplement
   * this function to create an object of the same class and the same event symboltable.
   * The StateNamesEnabled indicator shoul be copied, too.
   *
   * @return 
   *   new vGenerator 
   */
  virtual vGenerator* NewP(void) const = 0;

  /** @} doxygen group */

  /*****************************************
   *****************************************
   *****************************************
   *****************************************/

  /** @name Basic Maintenance */
  /** @{ doxygen group */


  /** 
   * Set the generator's name 
   *
   * @param rName
   *   Generator name
   */
  void Name(const std::string& rName);

  /** 
   * Get generator's name 
   *
   * @return 
   *   Name of generator
   */
  std::string Name(void) const;

  /** 
   * Get generator's unique Id (mId)
   *
   * @return 
   *   Id of Generator
   */
  Idx Id(void) const;



  /** @} doxygen group */

  /*****************************************
   *****************************************
   *****************************************
   *****************************************/

  /** @name Event Symboltable */
  /** @{ doxygen group */


  /**
   * Get Pointer to global EventSymbolTable. This is
   * a static member of SymbolTable and used as default
   * for all derived classes and instantiated objects.
   *
   * @return
   *   Pointer to global EventSymbolTable
   */
  static SymbolTable* GlobalEventSymbolTablep(void);


  /** @} doxygen group */

  /** @name Event Symboltable */
  /** @{ doxygen group */


  /**
   * Get Pointer to EventSymbolTable currently used
   * by this vGenerator. 
   *
   * @return
   *   Pointer to EventSymbolTable
   */
  SymbolTable* EventSymbolTablep(void) const;

  /**
   * Set EventSymbolTable to be used by this vGenerator. 
   * This function sets the reference to global the event symboltable. The current implementation
   * in derived classes clears the generator, future versions may implement a re-indexing.
   *
   * @param pSymTab
   *   Pointer to SymbolTable
   */
  virtual void EventSymbolTablep(SymbolTable* pSymTab);

  /**
   * Set EventSymbolTable as given by rOtherGen. 
   * This function sets the reference to the event symboltable. The current implementation
   * clears the generator, future versions may implement a re-indexing.
   *
   * @param rOtherGen
   *   other generator 
   *   
   */
  virtual void EventSymbolTablep(const vGenerator& rOtherGen);


  /**
   * EventSet with generator's EventSymbolTable (on stack).
   *
   * @return
   *   New empty EventSet on stack
   */
  EventSet NewEventSet(void) const;

  /**
   * EventSet with generator's EventSymbolTable (on heap). 
   *
   * @return
   *   Pointer to new empty EventSet on heap
   */
  EventSet* NewEventSetp(void) const;

  /**
   * Event index lookup 
   * 
   * @param rName
   *    Name of event to lookup
   *
   * @return 
   *   Valid index or 0 if non-existent
   */
  Idx EventIndex(const std::string& rName) const;

  /** 
   * Event name lookup 
   *
   * @param index
   *   Index of event to look up
   *
   * @return 
   *   Name or empty std::string if non-existent
   */
  std::string EventName(Idx index) const;

  /**
   * Set name for existing event
   *
   * @param index
   *   Event index
   * @param rName
   *   New name
   *
   * @exception Exception
   *   - index not found in EventSymbolMap (id 42)
   *   - name already associated with another index (id 44)
   *   - event does not exist in generator (id 89)
   */
  void EventName(Idx index, const std::string& rName);

  /** 
   * Create a new unique symbolic event name.
   * See also SymbolTable::UniqueSymbol().
   *
   * @param rName
   *   suggestion for new state name
   */
  std::string UniqueEventName(const std::string& rName) const;


  /** @} doxygen group */

  /*****************************************
   *****************************************
   *****************************************
   *****************************************/

  /** @name State Symboltable */
  /** @{ doxygen group */


  /**
   * Get StateSymbolTable. 
   *
   * @return
   *   ref to state-symboltable
   */
  const SymbolTable& StateSymbolTable(void) const;

  /**
   * Set StateSymbolTable.
   *
   * By convention, state names and indices are local to the
   * respective generator. It is most unlikely that you want to use
   * this function.
   *
   * @return
   *   Pointer to mpStateSymbolTable
   */
  void StateSymbolTable(const SymbolTable& rSymTab);

  /**
   * State index lookup. 
   *
   * @param rName
   *
   * @return
   *  Valid index (or 0 for non-existent)
   */
  Idx StateIndex(const std::string& rName) const;

  /** 
   * State name lookup 
   *
   * @param index
   *
   * @return name (or empty string if no such exists) 
   */
  std::string StateName(Idx index) const;

  /** 
   * Set name of state
   *
   * @param index
   *   Index
   * @param rName
   *   Name
   *
   * @exception Exception
   *   - name already associated with another index (id 44)
   *   - state does not exist in generator (id 90)
   */
  void StateName(Idx index, const std::string& rName);

  /**
   * Remove all names from generator's StateSymbolTable
   */
  void ClearStateNames(void);

  /**
   * Clear name for individual state
   *
   * @param index
   *   State index
   * @exception Exception
   *   - state does not exist in generator (id 90)
   */
  void ClrStateName(Idx index);

  /**
   * Clear name for individual  state
   *
   * @param rName
   *   State name
   */
  void ClrStateName(const std::string& rName);

  /**
   * Whether libFAUEDS functions are requested to generate state names.
   * Most libFAUDES functions introduce new states to a generator can be enabled
   * to also assign (more or less sensible) names to those states. This feature is
   * purely cosmetic and may be disabled.
   * 
   *
   * @return
   *   False if ClrStateNames() was called
   */
  bool StateNamesEnabled(void) const;

  /**
   * Enable/disable libFAUEDS functions to automatically generate state names.
   *
   * @param flag
   *   True enables statenames / false disables them
   */
  void StateNamesEnabled(bool flag);

  /**
   * Assign each state a default name bysed on its index.
   */
  void SetDefaultStateNames(void);

  /**
   * For all states without a symbolic name, assign a name 
   * based on suggested template and the index
   *
   * @param rTemplate
   *   basis for name generation
   */
  void EnforceStateNames(const std::string& rTemplate);

  /** 
   * Create a new unique symbolic state name.
   * See also SymbolTable::UniqueSymbol().
   *
   * @param rName
   *   suggestion for new state name
   */
  std::string UniqueStateName(const std::string& rName) const;


  /** @} doxygen group */

  /*****************************************
   *****************************************
   *****************************************
   *****************************************/

  /** @name Misc */
  /** @{ doxygen group */

  /**
   * Pretty printable event name for index (eg for debugging).
   *
   * @param index
   *   Event index
   *
   * @return
   *   std::string
   */
  std::string EStr(Idx index) const;

  /**
   * Return pretty printable state name for index (eg for debugging)
   *
   * @param index
   *   State index
   *
   * @return
   *   std::string
   */
  std::string SStr(Idx index) const;

  
  /**
   * Return pretty printable transition (eg for debugging)
   *
   * @param rTrans
   *   Transition
   *
   * @return
   *   std::string
   */
  std::string TStr(const Transition& rTrans) const;

  
  /**
   * Produce graphical representation of this generator.
   * This method calls the generator's DotWrite function and then processes the output
   * with the dot tool from graphiz package.  If no output format is given,
   * try to guess from filename extension.  See also ProcessDot().
   *
   * @param rFileName
   *   Name of output file
   * @param rOutFormat
   *   Graphics file format, eg "png", "jpg", "svg"
   * @param rDotExec
   *   path/name of executable
   * @exception Exception
   *   - IO errors (id 2)
   *   - error during systemcall for dot (id 3)
   */
  void GraphWrite(const std::string& rFileName, const std::string& rOutFormat="", 
    const std::string& rDotExec="dot");

  /** 
   * Order for sorting containers of generators 
   */
  bool operator < (const vGenerator& rOtherGen) const {
    return (mId < rOtherGen.mId);
  }

 
  /** @} doxygen group */

  /*****************************************
   *****************************************
   *****************************************
   *****************************************/

  /** @name Pure Virtual Function for Interface */
  /** @{ */


  /*
   * copy methods
   */

  virtual void Copy(vGenerator& rGen) const = 0; 
  virtual void Move(vGenerator& rGen) = 0; 
  virtual vGenerator& operator = (const vGenerator& rOtherGen) = 0; 

  // reason Vor version not being defined:
  // * would need vGenerator version in TaGenerator (to lazy)
  // * should have attribute transparent movetransition, moveevent etc
  // * should have Version to operate on this. 
  //virtual void Version(Idx version, vGenerator& rResGen) = 0;
  //virtual void Version(const std::string& rVersion, vGenerator& rResGen) = 0;



  /*
   * io
   */
  virtual void Write(void) const = 0;
  virtual  std::string ToString(void) const = 0;
  virtual void Write(const char* pFileName, 
		     std::ios::openmode openmode = std::ios::out|std::ios::trunc) = 0;
  virtual void Write(TokenWriter& rTw) const = 0;
  virtual void DWrite(void) const = 0;
  virtual void WriteAlphabet(void) const = 0;
  virtual std::string AlphabetToString(void) const = 0;
  virtual void WriteAlphabet(TokenWriter& rTw) const = 0;
  virtual void WriteStateSet(const StateSet& rStateSet) const = 0;
  virtual std::string StateSetToString(const StateSet& rStateSet) const = 0;
  virtual void WriteStateSet(TokenWriter& rTw, const StateSet& rStateSet) const = 0;
  virtual void DWriteStateSet(TokenWriter& rTw, const StateSet& rStateSet) const = 0;
  virtual std::string StatesToString(void) const = 0;
  virtual std::string MarkedStatesToString(void) const = 0;
  virtual std::string InitStatesToString(void) const = 0;
  virtual void WriteTransRel(void) const = 0;
  virtual std::string TransRelToString(void) const = 0;
  virtual void WriteTransRel(TokenWriter& rTw) const = 0;
  virtual void DWriteTransRel(TokenWriter& rTw) const = 0;
  virtual void WriteStatistics(void) const = 0;
  virtual std::string StatisticsToString(void) const = 0;
  virtual void WriteStatistics(TokenWriter& rTw) const = 0;
  virtual void DotWrite(const std::string& rFileName) = 0;
  virtual void DDotWrite(const std::string& rFileName) const = 0;

  virtual void Read(const char* pFileName) = 0;
  virtual void Read(TokenReader& rTr) = 0;
  virtual void ReadStateSet(TokenReader& rTr, const std::string& rLabel, StateSet& rStateSet) const = 0;


  /*
   * informative 
   */
  virtual bool Valid(void) = 0;
  virtual void Clear(void) = 0;
  virtual void ClearStates(void) = 0;
  virtual Idx Size(void) const = 0;
  virtual Idx AlphabetSize(void) const = 0;
  virtual Idx TransRelSize(void) const = 0;
  virtual Idx InitStatesSize(void) const = 0;
  virtual Idx MarkedStatesSize(void) const = 0;
  virtual bool Empty(void) const = 0;
  virtual bool AlphabetEmpty(void) const = 0;
  virtual bool TransRelEmpty(void) const = 0;
  virtual bool InitStatesEmpty(void) const = 0;
  virtual bool MarkedStatesEmpty(void) const = 0;
  virtual EventSet UsedEvents(void) const = 0;
  virtual EventSet UnusedEvents(void) const = 0;
  virtual EventSet ActiveEventSet(Idx x1) const = 0;
  virtual StateSet TransRelStateSpace(void) const = 0;
  virtual StateSet TransRelStateSpace(Idx x1) const = 0;


  /*
   * read access to core members
   */
  virtual EventSet::Iterator AlphabetBegin(void) const = 0;
  virtual EventSet::Iterator AlphabetEnd(void) const = 0;
  virtual bool ExistsEvent(Idx index) const = 0;
  virtual bool ExistsEvent(const std::string& rName) const = 0;
  virtual EventSet::Iterator FindEvent(Idx index) const = 0;
  virtual EventSet::Iterator FindEvent(const std::string& rName) const = 0;
  virtual const EventSet& Alphabet(void) const = 0;

  virtual StateSet::Iterator StatesBegin(void) const = 0;
  virtual StateSet::Iterator StatesEnd(void) const = 0;
  virtual bool ExistsState(Idx index) const = 0;
  virtual bool ExistsState(const std::string& name) const = 0;
  virtual StateSet::Iterator FindState(Idx index) const = 0;
  virtual StateSet::Iterator FindState(const std::string& rName) const = 0;
  virtual const StateSet& States(void) const = 0;

  virtual Idx InitState(void) const = 0;
  virtual StateSet::Iterator InitStatesBegin(void) const = 0;
  virtual StateSet::Iterator InitStatesEnd(void) const = 0;
  virtual bool ExistsInitState(Idx index) const = 0;
  virtual StateSet::Iterator FindInitState(Idx index) const = 0;
  virtual const StateSet& InitStates(void) const = 0;

  virtual StateSet::Iterator MarkedStatesBegin(void) const = 0;
  virtual StateSet::Iterator MarkedStatesEnd(void) const = 0;
  virtual bool ExistsMarkedState(Idx index) const = 0;
  virtual StateSet::Iterator FindMarkedState(Idx index) const = 0;
  virtual const StateSet& MarkedStates(void) const = 0;

  virtual TransSet::Iterator TransRelBegin(void) const = 0;
  virtual TransSet::Iterator TransRelEnd(void) const = 0;
  virtual TransSet::Iterator TransRelBegin(Idx x1) const = 0;
  virtual TransSet::Iterator TransRelEnd(Idx x1) const = 0;
  virtual TransSet::Iterator TransRelBegin(Idx x1, Idx ev) const = 0;
  virtual TransSet::Iterator TransRelEnd(Idx x1, Idx ev) const = 0;
  virtual TransSet::Iterator FindTransition(
					    const std::string& rX1, const std::string& rEv, const std::string& rX2) const = 0;
  virtual TransSet::Iterator FindTransition(Idx x1, Idx ev, Idx x2) const = 0;
  virtual TransSet::Iterator FindTransition(const Transition& rTrans) const = 0;
  virtual bool ExistsTransition(
				const std::string& rX1, const std::string& rEv, const std::string& rX2) const = 0;
  virtual bool ExistsTransition(Idx x1, Idx ev, Idx x2) const = 0;
  virtual bool ExistsTransition(const Transition& rTrans) const = 0;
  virtual const TransSet& TransRel(void) const = 0;
  virtual void TransRel(TransSetX1EvX2& res) const = 0;
  virtual void TransRel(TransSetEvX1X2& res) const = 0;
  virtual void TransRel(TransSetEvX2X1& res) const = 0;
  virtual void TransRel(TransSetX2EvX1& res) const = 0;
  virtual void TransRel(TransSetX2X1Ev& res) const = 0;
  virtual void TransRel(TransSetX1X2Ev& res) const = 0;

  /*
   * write access to core members
   */

  virtual bool InsEvent(Idx index) = 0;
  virtual Idx InsEvent(const std::string& rName) = 0;
  virtual void InsEvents(const EventSet& events) = 0;
  virtual bool DelEvent(Idx index) = 0;
  virtual bool DelEvent(const std::string& rName) = 0;
  virtual void DelEvents(const EventSet& rEvents) = 0;
  virtual bool DelEventFromAlphabet(Idx index) = 0;
  virtual void InjectAlphabet(const EventSet& newalphabet) = 0;

  virtual Idx InsState(void) = 0;
  virtual bool InsState(Idx index) = 0;
  virtual Idx InsState(const std::string& rName) = 0;
  virtual bool DelState(Idx index) = 0;
  virtual bool DelState(const std::string& rName) = 0;
  virtual void DelStates(const StateSet& rDelStates) = 0;
  virtual bool DelStateFromStates(Idx index) = 0;
  virtual StateSet::Iterator DelStateFromStates(StateSet::Iterator pos) = 0;
  virtual void InjectState(Idx index) = 0;
  virtual void InjectStates(const StateSet& rNewStates) = 0;

  virtual Idx InsInitState(void) = 0;
  virtual Idx InsInitState(const std::string& rName) = 0;
  virtual void SetInitState(Idx index) = 0;
  virtual void SetInitState(const std::string& rName) = 0;
  virtual void ClrInitState(Idx index) = 0;
  virtual void ClrInitState(const std::string& rName) = 0;
  virtual void ClrInitState(StateSet::Iterator pos) = 0;
  virtual void ClearInitStates(void) = 0; 
  virtual void InjectInitStates(const StateSet& rNewInitStates) = 0;

  virtual Idx InsMarkedState(void) = 0;
  virtual Idx InsMarkedState(const std::string& rName) = 0;
  virtual void SetMarkedState(Idx index) = 0;
  virtual void SetMarkedState(const std::string& rName) = 0;
  virtual void ClrMarkedState(Idx index) = 0;
  virtual void ClrMarkedState(const std::string& rName) = 0;
  virtual void ClrMarkedState(StateSet::Iterator pos) = 0;
  virtual void ClearMarkedStates(void) = 0;
  virtual void InjectMarkedStates(const StateSet& rNewMarkedStates) = 0;

  virtual bool SetTransition(Idx x1, Idx ev, Idx x2) = 0;
  virtual bool SetTransition(const std::string& rX1, const std::string& rEv, 
					   const std::string& rX2) = 0;
  virtual bool SetTransition(const Transition& rTransition) = 0;
  virtual void ClrTransition(Idx x1, Idx ev, Idx x2) = 0;
  virtual void ClrTransition(const Transition& rTrans) = 0;
  virtual TransSet::Iterator ClrTransition(TransSet::Iterator it) = 0;
  virtual void ClearTransRel(void) = 0;
  virtual void InjectTransition(const Transition& rTrans) = 0;
  virtual void InjectTransRel(const TransSet& newtransset) = 0;


  /*
   * reachability
   */
  virtual StateSet AccessibleSet(void) const = 0;
  virtual bool Accessible(void) = 0;
  virtual bool IsAccessible(void) const = 0;
  virtual StateSet CoaccessibleSet(void) const = 0;
  virtual bool Coaccessible(void) = 0;
  virtual bool IsCoaccessible(void) const = 0;
  virtual StateSet TrimSet(void) const = 0;
  virtual bool Trim(void) = 0;
  virtual bool IsTrim(void) const = 0;

  /*
   * misc
   */
  virtual bool IsDeterministic(void) const = 0;   
  virtual void SetMinStateIndexMap(void) = 0;
  virtual void ClearMinStateIndexMap(void)  = 0;
  virtual void MinimizeAlphabet(void) = 0;

  /** @} doxygen group */


 protected:

  /** Name of generator */
  std::string mMyName;

  /** Number of generator */
  Idx mId;

  /** Number of generator objects */
  static Idx msObjectCount;

  /** State symbol table (local per Generator)*/
  SymbolTable mStateSymbolTable;

  /** Pointer to State symbol table */
  SymbolTable* mpStateSymbolTable;

  /** Pointer to Event symbol table */
  SymbolTable* mpEventSymbolTable;

  /** No state names */
  bool mStateNamesEnabled;
    
};

} // namespace faudes



#define VGENERATOR_H
#endif

