/** @file nameset.h @brief Classes NameSet, TaNameSet */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef NAMESET_H

#include "definitions.h"
#include "exception.h"
#include "indexset.h"
#include "symboltable.h"
#include "tokenreader.h"
#include "tokenwriter.h"
#include <fstream>
#include <sstream>
#include <iostream>
#include <set>
#include <assert.h>

namespace faudes {

/** @addtogroup ContainerClasses */
/** @{ */

// Forward declaration for the attributed version of nameset
template<class Attr> class TaNameSet;


/**
 * Set of indices with symbolic names. 
 * In a NameSet, symbolic names are mandatory.
 * The class is derived from IndexSet and uses a pointer to a SymbolTable 
 * to maintain the symbolic names. The static SymbolTable is used as default,
 * which in the context of libfaudes becomes the global event symbol table.
 * It is an error to refer to an unknown symbolic name or to an index
 * with no name associated. When FAUDES_CHECKED is defined, an exception will
 * be thrown. It is also considered an error to relate two NameSets that refer to
 * different SymbolTables (using e.g. SetUnion).  
 *
 *
 * Since symbolic names are mandatory, file IO uses names rather than indices.
 * Technically, file IO is done by TaNameSet functions. This requires a copy and for
 * that reason may be reimplemented in a future versions.
 * The format is demonstrated by the following example
 * of a set with name "Alphabet" consisting of events "alpha", "beta" and "gamma":
 * \code
 * <Alphabet> 
 * "alpha" "beta" "gamma"
 * <\Alphabet> 
 * \endcode
 * Note that when reading a file, indices will be associated with the symbolic names
 * based on availability. Within one libfaudes session, each individual event will
 * be referred to by a unique index.
 */

class NameSet : public virtual IndexSet {

  /** 
   * We implement "protected privacy for template classes" by friendship.
   * This is used for the pragmatic implemention conversion constructors.
   */
  template<class Attr> friend class TaNameSet;

 public:
  /**
   * Constructor for NameSet referring to the static SymbolTable. 
   */
  NameSet(void);
         
  /**
   * Copy-constructor from other NameSet.
   * This also copies the SymbolTable reference, hence the new NameSet
   * will use the same SymbolTable as rOtherSet.
   *
   * @param rOtherSet
   *   Set to copy
   */
  NameSet(const NameSet& rOtherSet);


  /**
   * Constructor from file.
   * This constructor reads a NameSet from a file using the DoRead(TokenReader&, const std::string&)
   * function. The section is specified by rLabel and the static SymbolTable is used.
   *
   *
   * @param rFilename
   *   Name of file
   * @param rLabel
   *   Section for the set in the file; 
   */
  NameSet(const std::string& rFilename, const std::string& rLabel = "NameSet");

  /**
   * Return an empty NameSet with same SymbolTable as this set.
   *
   * @return 
   *   New empty NameSet
   */
  NameSet NewN() const;

  /**
   * Return a pointer to a NameSet with same Symboltable
   * assigned.
   *
   * @return
   *   Pointer to new empty NameSet on heap
   */
  NameSet* NewP() const;

  /**
   * Get Pointer mpSymbolTable.
   *
   * @return
   *   Pointer mpSymbolTable
   */
  SymbolTable* SymbolTablep(void) const;

  /**
   * Set SymbolTable reference.
   * This function sets the reference to the SymbolTable. The current implementation
   * clears the set, future versions may implement a re-indexing.
   *
   * @param pSymTab
   *   Pointer to SymbolTable
   */
  void SymbolTablep(SymbolTable* pSymTab);

  /**
   * Add an element by index. 
   * Index must be already known to the global SymbolTable. 
   *
   * @param index
   *   Index to add
   * @return 
   *   True, if element was new to set
   * @exception Exception
   *   - no symbolic name for index (id 65)
   */
  bool Insert(Idx index);

  /** 
   * Add an element by its symbolic name. If the name is unknown, 
   * a new index will be generated and recorded in the symboltable.
   * If the name is known, the corresponding index will be added to the set.
   *
   * @param rName
   *   Symbolic name of element to add
   *
   * @return 
   *   Index of (new) element
   */
  Idx Insert(const std::string& rName);

  /**
   * Inserts elements of rOtherSet.
   *
   * @param rOtherSet
   *   Other NameSet
   * @exception Exception
   *   - symboltable mismatch (id 67) 
   */
   void InsertSet(const NameSet& rOtherSet);

  /**
   * Delete element by index. The symbolic name is not removed from the SymbolTable.
   *
   * @param index
   *   Index
   * @return
   *   True if element did exist
   */
  bool Erase(Idx index);

  /**
   * Delete element by symbolic name. The symbolic name is not removed from the SymbolTable
   *
   * @param rName
   *   symbolic name
   * @return
   *    True if element did exist
   * @exception Exception
   *   - unknown symbolic name (id 66)
   */
  bool Erase(const std::string& rName);

  /**
   * Delete element by iterator. Symbolic nam is not removed from SymbolTable.
   *
   * @param pos
   *   NameSet::iterator
   * @return
   *  iterator to next element
   * @exception Exception
   *   - invalid iterator (id 62)
   */
  Iterator Erase(const Iterator& pos);

  /**
   * Erase elements specified by rOtherSet
   *
   * @param rOtherSet
   *   Other StateSet
   * @exception Exception
   *   - symboltable mismatch (id 67) 
   */
   void EraseSet(const NameSet& rOtherSet);

  /**
   * Set new name for existing index.
   * FAUDES_CHECKED checks if index exists in NameSet. 
   *
   * @param index
   *   Index to edit
   * @param rName
   *   New name
   *
   * @exception Exception
   *   - index not in this set (id 60)
   *   - index not found in SymbolTable (id 42)
   *   - name already associated with another index (id 44)
   */
  void SymbolicName(Idx index, const std::string& rName);

        
  /**
   * Set new name for existing name
   * FAUDES_CHECKED checks if the specified name exists in NameSet. 
   *
   * @param rName
   *   Symbolic name to edit
   * @param rNewName
   *   New name
   *
   * @exception Exception
   *   - symbolic name not in this set (id 60)
   *   - new name already associated with another index (id 44)
   */
  void SymbolicName(const std::string& rName, const std::string& rNewName);

  /**
   * Name lookup
   *
   * @param index
   *   Index to lookup
   *
   * @return
   *   Corresponding name or empty std::string if non-existent
   */
  std::string SymbolicName(Idx index) const;

  /**
   * Index lookup
   *
   * @param rName
   *   Symbolic name to look up
   *
   * @return 
   *   Corresponding index or 0 for non-existent
   */
  Idx Index(const std::string& rName) const;

  /**
   * Test existence of index
   *
   * @param index
   *   Index to test
   *
   * @return 
   *   True if index is in this set
   */
  bool Exists(Idx index) const;

  /** 
   * Test existence of name
   *
   * @param rName
   *   Symbolic name to test
   *
   * @return 
   *   True if index is in this set
   */
  bool Exists(const std::string& rName) const;

  /**
   * Find iterator for index. Returns either a valid iterator
   * or End() for non-existent.
   *
   * @param index
   *   Index to find
   *
   * @return
   *   NameSet::Iterator
   */
  NameSet::Iterator Find(Idx index) const;

  /**
   * Find iterator for name. Returns either a valid iterator
   * or End() for non-existent.
   *
   * @param rName
   *   Name to find
   *
   * @return
   *   NameSet::Iterator
   */
  NameSet::Iterator Find(const std::string& rName) const;


  /**
   * Set union operator
   *
   * @return 
   *   Union Set
   *
   * @exception Exception
   *   - symboltable mismatch (id 67) 
   */
  NameSet operator + (const NameSet& rOtherSet) const;

  /**
   * Set difference operator
   *
   * @return 
   *   Difference NameSet
   *
   * @exception Exception
   *   - symboltable mismatch (id 67) 
   */
  NameSet operator - (const NameSet& rOtherSet) const;

  /**
   * Set intersection operator
   *
   * @return 
   *   Intersection NameSet
   *
   * @exception Exception
   *   - symboltable mismatch (id 67) 
   */
   NameSet operator * (const NameSet& rOtherSet) const;


  /** Test for equality  */
  bool operator == (const NameSet& rOtherSet) const;

  /** Test for unequality  */
  bool operator != (const NameSet& rOtherSet) const;

  /** Test for subset  */
  bool operator <= (const NameSet& rOtherSet) const;

  /** Test for superset */
  bool operator >= (const NameSet& rOtherSet) const;


  /**
   * Return pretty printable symbolic name for index.
   * Primary meant for debugging messages.
   *
   * @param index
   *   Index to print
   *
   * @return
   *   String
   */
  std::string Str(Idx index) const;


 protected:

  /** Pointer to local SymbolTable */
  SymbolTable* mpSymbolTable;

  /**
   * Write to TokenWriter, see TBaseSet for public wrappers
   *
   * @param tw
   *   Reference to TokenWriter
   * @param rLabel
   *   Label of the section to write
   *
   * @exception Exception
   *   - IO errors (id 2)
   */
  void DoWrite(TokenWriter& tw, const std::string& rLabel) const;
        
  /** Write debug info to TokenWriter, see TBaseSet for public wrapper.
   * The debug version writes a format that includes symbolic names and indices. 
   *
   * @param tw
   *   Reference to TokenWriter
   *
   * @exception Exception 
   *   - IO errors (id 2)
   */
  void DoDWrite(TokenWriter& tw) const;

  /**
   * Read from TokenReader, see TBaseSet for public wrappers. 
   * Uses TaNameSet::Read(TokenReader&,const std::string&) with void attributes. 
   * It is an error if the file contains an index (id 52).
   *
   * @param tr
   *   Reference to TokenReader
   * @param rLabel
   *   Label to read
   *
   * @exception Exception
   *   - IO errors (id 1)
   *   - token mismatch (id 50, 51, 52)
   */
  void DoRead(TokenReader& tr, const std::string& rLabel = "NameSet");


};


/* convenience typedef for eventset */
typedef  NameSet EventSet;  


/**
 * Set of indices with symbolic names and attributes. 
 *
 * This class is derived from NameSet and TaIndexSet to inherit
 * both extra features on top of IndexSet.
 *
 * The file format is demonstrated by the following example
 * of a set "Alphabet" consisting of events "alpha", "beta" and "gamma"
 * with "gamma" having some attribute (see eg AtributeFlags)
 * \code
 * <Alphabet> 
 * "alpha" 
 * "beta" 
 * "gamma" 0x0f
 * <\Alphabet> 
 * \endcode
 * As with TaIndexSet, reading a file ignores unknown attributes. Thus, the above example
 * may also be read as IndexSet.
 */



template<class Attr>
class TaNameSet : public NameSet , public TaIndexSet<Attr> {

  /** 
   * We implement "protected privacy for template classes" by friendship.
   * This is used for the pragmatic implemention conversion constructors.
   */
  template<class A> friend class TaNameSet;


 public:
  /**
   * Constructor for NameSet referring to the static SymbolTable. 
   */
  TaNameSet(void);
         
  /**
   * Copy-constructor from other TaNameSet (incl attributes and symboltable) 
   *
   * @param rOtherSet
   *   Set to copy
   */
  TaNameSet(const TaNameSet& rOtherSet);

  /**
   * Constructor from NameSet (sets default attributes, same symboltable) 
   *
   * @param rOtherSet
   *   Set to copy
   */
  TaNameSet(const NameSet& rOtherSet);

  /**
   * Constructor from file.
   * This constructor reads a NameSet from a file using the DoRead(TokenReader&, const std::String&)
   * function. The section is specified by rLabel and the static SymbolTable is used.
   *
   * @param rFilename
   *   Name of File
   * @param rLabel
   *   Section for the set in the file; 
   */
  TaNameSet(const std::string& rFilename, const std::string& rLabel = "NameSet");

  /**
   * Virtual destructor
   */
   virtual ~TaNameSet(void) {}

  /**
   * Return a NameSet with same symboltable as this set.
   *
   * @return 
   *   New empty TaNameSet
   */
  TaNameSet NewN() const;

  /**
   * Return an empty NameSet with same SymbolTable as this set.
   *
   * @return
   *   Pointer to new empty NameSet on heap
   */
  TaNameSet* NewP() const;

  /**
   * Copy this set to any nameset, ignore attributes
   *
   * @param rOtherSet
   *   NameSet with no attributes
   */
  template<class OtherAttr>
  void CopyWithoutAttributes(TaNameSet<OtherAttr>& rOtherSet) const;


  /**
   * Add an element by index. 
   * Index must be already known to the global SymbolTable. If the element already
   * exists in the set, the attribute is maintained. Otherwise, the element
   * is inserted with default attribute.
   *
   * @param index
   *   Index to add
   * @return 
   *   True, if element was new to set
   * @exception Exception
   *   - no symbolic name for index (id 65)
   */
  bool Insert(Idx index);


  /**
   * Add an element by index incl. attribute 
   *
   * @param index
   *   Index to add
   * @param rAttr
   *   Attribute to add
   * @return 
   *   True, if index was new to set
   * @exception Exception
   *   - no symbolic name for index (id 65)
   *
   */
  bool Insert(Idx index, const Attr& rAttr);

  /** 
   * Add an element by its  symbolic name. If the name is unknown, 
   * a new index will be generated and recorded in the symboltable.
   * If the name is known, the corresponding index will be added to the set.
   * If the element already exists in the set, the attribute is maintained. 
   * Otherwise, the element is inserted with default attribute.
   *
   * @param rName
   *   symbolic name of element to add
   *
   * @return 
   *   Index of (new) element
   */
  Idx Insert(const std::string& rName);

  /** 
   * Add an element by its symbolic name.
   * If the name is unknown, 
   * a new index will be generated and recorded in the symboltable.
   * If the name is known, the corresponding index will be added to the set.
   *
   * @param rName
   *   symbolic name of element to add
   * @param rAttr
   *   Attribute
   *
   * @return 
   *   Index of (new) element
   */
  Idx Insert(const std::string& rName, const Attr& rAttr);

  /**
   * Inserts elements of rOtherSet.
   * Attributes of this set are maintained, newly inserted elements have default attribute.
   *
   * @param rOtherSet
   *   Other StateSet
   * @exception Exception 
   *   - symboltable mismatch (id 67)
   */
   void InsertSet(const NameSet& rOtherSet);

  /**
   * Inserts elements of rOtherSet incl attributes
   * Attributes of this set are maintained, new elements are inserted with attribute.
   *
   * @param rOtherSet
   *   Other StateSet
   */
   void InsertSet(const TaNameSet& rOtherSet);

  /**
   * Delete element by index. Symbolic name is not removed from SymbolTable.
   *
   * @param index
   *   Index to delete
   * @return
   *   True if element did exist
   * 
   */
  bool Erase(Idx index);

  /**
   * Delete element by symbolic name. Symbolic name is not removed from SymbolTable
   *
   * @param rName
   *   Symbolic name of element to dlete
   * @return
   *    True if element did exist
   * @exception Exception 
   *   - name not found in Symboltable (id 66)
   */
  bool Erase(const std::string& rName);

  /**
   * Delete element by iterator. Symbolic nam is not removed from SymbolTable.
   *
   * @param pos
   *   TaNameSet::iterator
   * @return
   *   Iterator to next element
   * @exception Exception 
   *   - invalid iterator (id 62)
   */
  Iterator Erase(const Iterator& pos);

  /**
   * Erase elements indicated by rOtherSet
   *
   * @exception Exception 
   *   - symboltable mismatch (id 67)
   *
   * @param rOtherSet
   *   Other StateSet
   */
   void EraseSet(const NameSet& rOtherSet);

  /**
   * Erase elements indicated by rOtherSet.
   *
   * @exception Exception 
   *   - symboltable mismatch (id 67)
   *
   * @param rOtherSet
   *   Other StateSet
   */
   void EraseSet(const TaNameSet& rOtherSet);

  /**
   * Return pretty printable symbolic name for index.
   * Primary meant for debugging messages.
   *
   * @param index
   *   Index to print
   *
   * @return
   *   String
   */
  std::string Str(Idx index) const;

 protected:

  /**
   * Write to TokenWriter, see TBaseSet for public wrappers.
   *
   * @param tw
   *   Reference to TokenWriter
   * @param rLabel
   *   Label for set in file
   *
   * @exception Exception
   *   - IO errors (id 2)
   */
  void DoWrite(TokenWriter& tw, const std::string& rLabel) const;
        

  /** 
   * Write debug info to TokenWriter, see TBaseSet for public wrappers.
   * The debug version writes a format that includes symbolic names and indices. 
   *
   * @param tw
   *   Reference to TokenWriter
   *
   * @exception Exception 
   *   - IO errors (id 2)
   */
  void DoDWrite(TokenWriter& tw) const;

  /**
   * Read from TokenReader, see TBaseSet for public wrappers.
   * It is an error if the file contains an index (id 52).
   *
   * @param tr
   *   Reference to TokenReader
   * @param rLabel
   *   Label to read
   *
   * @exception Exception
   *   - IO errors (id 1)
   *   - token mismatch (id 50, 51, 52)
   */
  void DoRead(TokenReader& tr, const std::string& rLabel = "NameSet");


};


/** Convenience Macro */
#define TaEventSet TaNameSet  

/** @} doxygen group */

/*
*************************************************************************************
*************************************************************************************
 Implementation
*************************************************************************************
*************************************************************************************
*/



// empty constructor
template<class Attr>
TaNameSet<Attr>::TaNameSet(void) {
  FD_DC("TaNameSet(" << this << ")::TaNameSet()");
  mpSymbolTable = SymbolTable::GlobalEventSymbolTablep();
  IndexSet::Name("NameSet");
}

// constructor form other nameset
template<class Attr>
TaNameSet<Attr>::TaNameSet(const TaNameSet& rOtherSet) :
  IndexSet(),                     // todo: init in diamond shape inheritance
  NameSet(rOtherSet),
  TaIndexSet<Attr>(rOtherSet)
{
  FD_DC("TaNameSet(" << this << ")::TaNameSet(rOtherSet " << &rOtherSet << ")");
  TaIndexSet<Attr>::mAttributeMap=rOtherSet.mAttributeMap;
}

// constructor form other nameset
template<class Attr>
TaNameSet<Attr>::TaNameSet(const NameSet& rOtherSet) :
  NameSet(rOtherSet),
  TaIndexSet<Attr>(rOtherSet)
{
  FD_DC("TaNameSet(" << this << ")::TaNameSet(rOtherSet " << &rOtherSet << ")");
}


// read file constructor
template<class Attr>
TaNameSet<Attr>::TaNameSet(const std::string& rFilename, const std::string& rLabel) {
  FD_DC("TaNameSet(" << this << ")::TaNameSet(" << rFilename << ")");
  mpSymbolTable = SymbolTable::GlobalEventSymbolTablep();
  Read(rFilename, rLabel);
}

// NewN()
template<class Attr>
TaNameSet<Attr> TaNameSet<Attr>::NewN(void) const {
  TaNameSet res;
  res.mpSymbolTable=mpSymbolTable;
  return res;
}

// NewP()
template<class Attr>
TaNameSet<Attr>* TaNameSet<Attr>::NewP(void) const {
  TaNameSet* res = new TaNameSet();
  res.mpSymbolTable=mpSymbolTable;
  return res;
}

// CopyWithoutAttributes(rOtherSet)
template<class Attr>
template<class OtherAttr>
void TaNameSet<Attr>::CopyWithoutAttributes(TaNameSet<OtherAttr>& rOtherSet) const {
  rOtherSet.Clear();
  rOtherSet.mMyName=TaIndexSet<Attr>::mMyName;
  rOtherSet.mIndexSet=TaIndexSet<Attr>::mIndexSet;
  rOtherSet.mpSymbolTable=mpSymbolTable;
}  

// DoWrite(tw, rLabel)
template<class Attr>
void TaNameSet<Attr>::DoWrite(TokenWriter& tw, const std::string& rLabel) const {
  std::string label=rLabel;
  if(label=="") label=Name(); 
  tw.WriteBegin(label);
  Token token;
  Iterator it;
  for (it = TaIndexSet<Attr>::Begin(); it != TaIndexSet<Attr>::End(); ++it) {
#ifdef FAUDES_CHECKED
    if (SymbolicName(*it) == "") {
      FD_ERR("TaNameSet::Write(): "
	   << "index " << *it << " not in SymbolTable. aborting...");
      abort();
    }
#endif
    token.SetString(SymbolicName(*it));
    tw << token;
    TaIndexSet<Attr>::Attribute(*it).Write(tw);
  }
  tw.WriteEnd(label);
}


// DoDWrite(tw)
template<class Attr>
void TaNameSet<Attr>::DoDWrite(TokenWriter& tw) const {
  Token token;
  Iterator it;
  tw.WriteBegin(IndexSet::Name());
  for (it = TaIndexSet<Attr>::Begin(); it != TaIndexSet<Attr>::End(); ++it) {
    tw << Str(*it);
    TaIndexSet<Attr>::Attribute(*it).Write(tw);
  }
  tw.WriteEnd(IndexSet::Name());
}


// DoRead(rTr, rLabel)
template<class Attr>
void TaNameSet<Attr>::DoRead(TokenReader& rTr, const std::string& rLabel) {
  FD_DC("TaNameSet(" << this << ")::DoRead(tr," << rLabel << ") with symtab " << NameSet::mpSymbolTable);
  std::string name;
  Attr attribute;
  AttributeVoid vattr;
  IndexSet::Name(rLabel);
  Token token;
  TaIndexSet<Attr>::Clear();

  rTr.SeekBegin(rLabel); 
  while(rTr.Peek(token)) {
    // break on end
    if (token.Type() == Token::End) {
      break;
    }
    // read by name
    if (token.Type() == Token::String) {
      rTr >> token;
      FD_DC("TaNameSet(" << this << ")::DoRead(..): inserting element \"" 
	    << token.StringValue() << "\"");
      name=token.StringValue();
      // read my attribute
      attribute.Read(rTr);
      // skip other attributes
      vattr.Read(rTr);
      // insert element
      Insert(name,attribute);
      continue;
    }
    // cannot process token
    std::stringstream errstr;
    errstr << "Invalid token" << rTr.FileLine();
    throw Exception("TaNameSet::DoRead", errstr.str(), 50);
  }
  rTr.SeekEnd(rLabel);
}



// Insert(index)
template<class Attr>
bool TaNameSet<Attr>::Insert(Idx index) { 
#ifdef FAUDES_CHECKED
  if(!mpSymbolTable->Exists(index)) {
    std::stringstream errstr;
    errstr << "index " << index << " has no symbolic name" << std::endl;
    throw Exception("TaNameSet::Insert", errstr.str(), 65);
  }
#endif
  return TaIndexSet<Attr>::Insert(index);
}

        
// Insert(index,attr)
template<class Attr>
bool TaNameSet<Attr>::Insert(Idx index, const Attr& attr) { 
#ifdef FAUDES_CHECKED
  if(!mpSymbolTable->Exists(index)) {
    std::stringstream errstr;
    errstr << "index " << index << " has no symbolic name" << std::endl;
    throw Exception("TaNameSet::Insert", errstr.str(), 65);
  }
#endif
  return TaIndexSet<Attr>::Insert(index,attr);
}

        
// Insert(rName)
template<class Attr>
Idx TaNameSet<Attr>::Insert(const std::string& rName) {
  FD_DC("TaNameSet(" << this << ")::Insert(" << rName <<")");
  Idx index= NameSet::Insert(rName); // automatic: keep attribute if exists
  return index;
}

// Insert(rName, attr)
template<class Attr>
Idx TaNameSet<Attr>::Insert(const std::string& rName, const Attr& attr) {
  FD_DC("TaNameSet(" << this << ")::Insert(" << rName <<")");
  Idx index= NameSet::Insert(rName);
  TaIndexSet<Attr>::Attribute(index,attr);
  return index;
}

// InsertSet(set)
template<class Attr>
void TaNameSet<Attr>::InsertSet(const NameSet& rOtherSet) {
  FD_DC("TaNameSet(" << this << ")::InsertSet(" << rOtherSet.ToString() << ")");
#ifdef FAUDES_CHECKED
  if(rOtherSet.mpSymbolTable!=mpSymbolTable) {
    std::stringstream errstr;
    errstr << "symboltable mismach aka not implemented" << std::endl;
    throw Exception("TaNameSet::InsertSet", errstr.str(), 67);
  }
#endif
  NameSet::InsertSet(rOtherSet);
}

// InsertSet(set)
template<class Attr>
void TaNameSet<Attr>::InsertSet(const TaNameSet& rOtherSet) {
  FD_DC("TaNameSet(" << this << ")::InsertSet(" << rOtherSet.ToString() << ")");
#ifdef FAUDES_CHECKED
  if(rOtherSet.mpSymbolTable!=mpSymbolTable) {
    std::stringstream errstr;
    errstr << "symboltable mismach aka not implemented" << std::endl;
    throw Exception("TaNameSet::InsertSet", errstr.str(), 200);
  }
#endif
  TaIndexSet<Attr>::InsertSet(rOtherSet);
}
		
// Erase(index)
template<class Attr>
bool TaNameSet<Attr>::Erase(Idx index) {
  FD_DC("TaNameSet(" << this << ")::Erase(" << index <<")");
  return TaIndexSet<Attr>::Erase(index);
}

// Erase(rName)
template<class Attr>
bool TaNameSet<Attr>::Erase(const std::string& rName) {
  FD_DC("TaNameSet(" << this << ")::Erase(" << rName <<")");
  Idx index = mpSymbolTable->Index(rName);
#ifdef FAUDES_CHECKED
  if (index == 0) {
    std::stringstream errstr;
    errstr << "name \"" << rName << "\" not found in TaNameSet" << std::endl;
    throw Exception("TaNameSet::Erase", errstr.str(), 60);
  }
#endif
  return TaIndexSet<Attr>::Erase(index);
}

// Erase(pos)
template<class Attr>
typename TaNameSet<Attr>::Iterator TaNameSet<Attr>::Erase(const Iterator& pos) {
  return TaIndexSet<Attr>::Erase(pos);
}

// EraseSet(set)
template<class Attr>
void TaNameSet<Attr>::EraseSet(const NameSet& rOtherSet) {
  FD_DC("TaNameSet(" << this << ")::EraseSet(" << rOtherSet.ToString() << ")");
#ifdef FAUDES_CHECKED
  if(rOtherSet.mpSymbolTable!=mpSymbolTable) {
    std::stringstream errstr;
    errstr << "symboltable mismach aka not implemented" << std::endl;
    throw Exception("TaNameSet::EraseSet", errstr.str(), 67);
  }
#endif
  TaIndexSet<Attr>::EraseSet(rOtherSet);
}

// EraseSet(set)
template<class Attr>
void TaNameSet<Attr>::EraseSet(const TaNameSet& rOtherSet) {
  FD_DC("TaNameSet(" << this << ")::EraseSet(" << rOtherSet.ToString() << ")");
#ifdef FAUDES_CHECKED
  if(rOtherSet.mpSymbolTable!=mpSymbolTable) {
    std::stringstream errstr;
    errstr << "symboltable mismatch aka not implemented" << std::endl;
    throw Exception("TaNameSet::EraseSet", errstr.str(), 67);
  }
#endif
  TaIndexSet<Attr>::EraseSet(rOtherSet);
}
		
// Str()
template<class Attr>
std::string TaNameSet<Attr>::Str(Idx index) const {
  return NameSet::Str(index);
}



} // namespace faudes

#define NAMESET_H
#endif 

